# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import time

import numpy as np

from pymor.core.base import BasicObject, abstractmethod
from pymor.core.exceptions import ExtensionError
from pymor.core.logger import getLogger
from pymor.parallel.dummy import dummy_pool
from pymor.parallel.interface import RemoteObject


def weak_greedy(surrogate, training_set, atol=None, rtol=None, max_extensions=None, pool=None):
    """Weak greedy basis generation algorithm :cite:`BCDDPW11`.

    This algorithm generates an approximation basis for a given set of vectors
    associated with a training set of parameters by iteratively evaluating a
    :class:`surrogate <WeakGreedySurrogate>` for the approximation error on
    the training set and adding the worst approximated vector (according to
    the surrogate) to the basis.

    The constructed basis is extracted from the surrogate after termination
    of the algorithm.

    Parameters
    ----------
    surrogate
        An instance of :class:`WeakGreedySurrogate` representing the surrogate
        for the approximation error.
    training_set
        The set of parameter samples on which to perform the greedy search.
    atol
        If not `None`, stop the algorithm if the maximum (estimated) error
        on the training set drops below this value.
    rtol
        If not `None`, stop the algorithm if the maximum (estimated)
        relative error on the training set drops below this value.
    max_extensions
        If not `None`, stop the algorithm after `max_extensions` extension
        steps.
    pool
        If not `None`, a |WorkerPool| to use for parallelization. Parallelization
        needs to be supported by `surrogate`.

    Returns
    -------
    Dict with the following fields:

        :max_errs:               Sequence of maximum estimated errors during the greedy run.
        :max_err_mus:            The parameters corresponding to `max_errs`.
        :extensions:             Number of performed basis extensions.
        :time:                   Total runtime of the algorithm.
    """
    logger = getLogger('pymor.algorithms.greedy.weak_greedy')
    training_set = list(training_set)
    logger.info(f'Started greedy search on training set of size {len(training_set)}.')

    tic = time.perf_counter()
    if not training_set:
        logger.info('There is nothing else to do for an empty training set.')
        return {'max_errs': [], 'max_err_mus': [], 'extensions': 0,
                'time': time.perf_counter() - tic}

    if pool is None:
        pool = dummy_pool
    elif pool is not dummy_pool:
        logger.info(f'Using pool of {len(pool)} workers for parallel greedy search.')

    # Distribute the training set evenly among the workers.
    if pool:
        training_set = pool.scatter_list(training_set)

    extensions = 0
    max_errs = []
    max_err_mus = []

    while True:
        with logger.block('Estimating errors ...'):
            max_err, max_err_mu = surrogate.evaluate(training_set)
            max_errs.append(max_err)
            max_err_mus.append(max_err_mu)

        logger.info(f'Maximum error after {extensions} extensions: {max_err} (mu = {max_err_mu})')

        if atol is not None and max_err <= atol:
            logger.info(f'Absolute error tolerance ({atol}) reached! Stopping extension loop.')
            break

        if rtol is not None and max_err / max_errs[0] <= rtol:
            logger.info(f'Relative error tolerance ({rtol}) reached! Stopping extension loop.')
            break

        with logger.block(f'Extending surrogate for mu = {max_err_mu} ...'):
            try:
                surrogate.extend(max_err_mu)
            except ExtensionError:
                logger.info('Extension failed. Stopping now.')
                break
            extensions += 1

        logger.info('')

        if max_extensions is not None and extensions >= max_extensions:
            logger.info(f'Maximum number of {max_extensions} extensions reached.')
            break

    tictoc = time.perf_counter() - tic
    logger.info(f'Greedy search took {tictoc} seconds')
    return {'max_errs': max_errs, 'max_err_mus': max_err_mus, 'extensions': extensions,
            'time': tictoc}


class WeakGreedySurrogate(BasicObject):
    """Surrogate for the approximation error in :func:`weak_greedy`."""

    @abstractmethod
    def evaluate(self, mus, return_all_values=False):
        """Evaluate the surrogate for given parameters.

        Parameters
        ----------
        mus
            List of parameters for which to estimate the approximation
            error. When parallelization is used, `mus` can be a |RemoteObject|.
        return_all_values
            See below.

        Returns
        -------
        If `return_all_values` is `True`, an |array| of the estimated errors.
        If `return_all_values` is `False`, the maximum estimated error as first
        return value and the corresponding parameter as second return value.
        """
        pass

    @abstractmethod
    def extend(self, mu):
        pass


def rb_greedy(fom, reductor, training_set, use_error_estimator=True, error_norm=None,
              atol=None, rtol=None, max_extensions=None, extension_params=None, pool=None):
    """Weak Greedy basis generation using the RB approximation error as surrogate.

    This algorithm generates a reduced basis using the :func:`weak greedy <weak_greedy>`
    algorithm :cite:`BCDDPW11`, where the approximation error is estimated from computing
    solutions of the reduced order model for the current reduced basis and then estimating
    the model reduction error.

    Parameters
    ----------
    fom
        The |Model| to reduce.
    reductor
        Reductor for reducing the given |Model|. This has to be
        an object with a `reduce` method, such that `reductor.reduce()`
        yields the reduced model, and an `exted_basis` method,
        such that `reductor.extend_basis(U, copy_U=False, **extension_params)`
        extends the current reduced basis by the vectors contained in `U`.
        For an example see :class:`~pymor.reductors.coercive.CoerciveRBReductor`.
    training_set
        The training set of |Parameters| on which to perform the greedy search.
    use_error_estimator
        If `False`, exactly compute the model reduction error by also computing
        the solution of `fom` for all |parameter values| of the training set.
        This is mainly useful when no estimator for the model reduction error
        is available.
    error_norm
        If `use_error_estimator` is `False`, use this function to calculate the
        norm of the error. If `None`, the Euclidean norm is used.
    atol
        See :func:`weak_greedy`.
    rtol
        See :func:`weak_greedy`.
    max_extensions
        See :func:`weak_greedy`.
    extension_params
        `dict` of parameters passed to the `reductor.extend_basis` method.
        If `None`, `'gram_schmidt'` basis extension will be used as a default
        for stationary problems (`fom.solve` returns `VectorArrays` of length 1)
        and `'pod'` basis extension (adding a single POD mode) for instationary
        problems.
    pool
        See :func:`weak_greedy`.

    Returns
    -------
    Dict with the following fields:

        :rom:                    The reduced |Model| obtained for the
                                 computed basis.
        :max_errs:               Sequence of maximum errors during the greedy run.
        :max_err_mus:            The parameters corresponding to `max_errs`.
        :extensions:             Number of performed basis extensions.
        :time:                   Total runtime of the algorithm.
    """
    surrogate = RBSurrogate(fom, reductor, use_error_estimator, error_norm, extension_params, pool or dummy_pool)

    result = weak_greedy(surrogate, training_set, atol=atol, rtol=rtol, max_extensions=max_extensions, pool=pool)
    result['rom'] = surrogate.rom

    return result


class RBSurrogate(WeakGreedySurrogate):
    """Surrogate for the :func:`weak_greedy` error used in :func:`rb_greedy`.

    Not intended to be used directly.
    """

    def __init__(self, fom, reductor, use_error_estimator, error_norm, extension_params, pool):
        self.__auto_init(locals())
        if use_error_estimator:
            self.remote_fom, self.remote_error_norm, self.remote_reductor = None, None, None
        else:
            self.remote_fom, self.remote_error_norm, self.remote_reductor = \
                pool.push(fom), pool.push(error_norm), pool.push(reductor)
        self.rom = None

    def evaluate(self, mus, return_all_values=False):
        if self.rom is None:
            with self.logger.block('Reducing ...'):
                self.rom = self.reductor.reduce()

        if not isinstance(mus, RemoteObject):
            mus = self.pool.scatter_list(mus)

        result = self.pool.apply(_rb_surrogate_evaluate,
                                 rom=self.rom,
                                 fom=self.remote_fom,
                                 reductor=self.remote_reductor,
                                 mus=mus,
                                 error_norm=self.remote_error_norm,
                                 return_all_values=return_all_values)
        if return_all_values:
            return np.hstack(result)
        else:
            errs, max_err_mus = list(zip(*result))
            max_err_ind = np.argmax(errs)
            return errs[max_err_ind], max_err_mus[max_err_ind]

    def extend(self, mu):
        with self.logger.block(f'Computing solution snapshot for mu = {mu} ...'):
            U = self.fom.solve(mu)
        with self.logger.block('Extending basis with solution snapshot ...'):
            extension_params = self.extension_params
            if len(U) > 1:
                if extension_params is None:
                    extension_params = {'method': 'pod'}
                else:
                    extension_params.setdefault('method', 'pod')
            self.reductor.extend_basis(U, copy_U=False, **(extension_params or {}))
            if not self.use_error_estimator:
                self.remote_reductor = self.pool.push(self.reductor)
        with self.logger.block('Reducing ...'):
            self.rom = self.reductor.reduce()


def _rb_surrogate_evaluate(rom=None, fom=None, reductor=None, mus=None, error_norm=None, return_all_values=False):
    if not mus:
        if return_all_values:
            return []
        else:
            return -1., None

    if fom is None:
        errors = [rom.estimate_error(mu) for mu in mus]
    elif error_norm is not None:
        errors = [error_norm(fom.solve(mu) - reductor.reconstruct(rom.solve(mu))) for mu in mus]
    else:
        errors = [(fom.solve(mu) - reductor.reconstruct(rom.solve(mu))).norm() for mu in mus]
    # most error_norms will return an array of length 1 instead of a number,
    # so we extract the numbers if necessary
    errors = [x[0] if hasattr(x, '__len__') else x for x in errors]
    if return_all_values:
        return errors
    else:
        max_err_ind = np.argmax(errors)
        return errors[max_err_ind], mus[max_err_ind]



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)
import warnings

from pymor.analyticalproblems.domaindescriptions import RectDomain
from pymor.analyticalproblems.elliptic import StationaryProblem
from pymor.analyticalproblems.functions import ConstantFunction, LincombFunction, BitmapFunction
from pymor.core.defaults import defaults
from pymor.parameters.functionals import ProjectionParameterFunctional
from pymor.tools.io import safe_temporary_filename


@defaults('font_name')
def text_problem(text='pyMOR', font_name=None):
    import numpy as np
    from PIL import Image, ImageDraw

    font = _get_font(font_name)
    size = font.getsize(text)  # compute width and height of rendered text
    size = (size[0] + 20, size[1] + 20)  # add a border of 10 pixels around the text

    def make_bitmap_function(char_num):
        # we need to generate a BitmapFunction for each character
        img = Image.new('L', size)  # create new Image object of given dimensions
        d = ImageDraw.Draw(img)  # create ImageDraw object for the given Image

        # in order to position the character correctly, we first draw all characters from the first
        # up to the wanted character
        d.text((10, 10), text[:char_num + 1], font=font, fill=255)

        # next we erase all previous character by drawing a black rectangle
        if char_num > 0:
            d.rectangle(((0, 0), (font.getsize(text[:char_num])[0] + 10, size[1])), fill=0, outline=0)

        # open a new temporary file
        # after leaving this 'with' block, the temporary file is automatically deleted
        with safe_temporary_filename(name='letter_bitmap.png') as f:
            img.save(f, format='png')
            return BitmapFunction(f, bounding_box=[(0, 0), size], range=[0., 1.])

    # create BitmapFunctions for each character
    dfs = [make_bitmap_function(n) for n in range(len(text))]

    # create an indicator function for the background
    background = ConstantFunction(1., 2) - LincombFunction(dfs, np.ones(len(dfs)))

    # form the linear combination
    dfs = [background] + dfs
    coefficients = [1] + [ProjectionParameterFunctional('diffusion', len(text), i) for i in range(len(text))]
    diffusion = LincombFunction(dfs, coefficients)

    return StationaryProblem(
        domain=RectDomain(dfs[1].bounding_box, bottom='neumann'),
        neumann_data=ConstantFunction(-1., 2),
        diffusion=diffusion,
        parameter_ranges=(0.1, 1.)
    )


def _get_font(font_name):
    from PIL import ImageFont
    font_list = [font_name] if font_name else ['DejaVuSansMono.ttf', 'VeraMono.ttf', 'UbuntuMono-R.ttf', 'Arial.ttf']
    for filename in font_list:
        try:
            return ImageFont.truetype(filename, 64)  # load some font from file of given size
        except (OSError, IOError):
            pass
    warnings.warn(f'Falling back to pillow default font. Could not load any of {font_list}',
                  ResourceWarning, stacklevel=1)
    return ImageFont.load_default()



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""IRKA-type reductor for |SecondOrderModels|."""

from pymor.models.iosys import SecondOrderModel
from pymor.reductors.h2 import GenericIRKAReductor, IRKAReductor
from pymor.reductors.interpolation import SOBHIReductor


class SORIRKAReductor(GenericIRKAReductor):
    """SOR-IRKA reductor.

    Parameters
    ----------
    fom
        The full-order |SecondOrderModel| to reduce.
    mu
        |Parameter values|.
    """

    def __init__(self, fom, mu=None):
        assert isinstance(fom, SecondOrderModel)
        super().__init__(fom, mu=mu)

    def reduce(self, rom0_params, tol=1e-4, maxit=100, num_prev=1,
               force_sigma_in_rhp=False, projection='orth', conv_crit='sigma',
               compute_errors=False, irka_options=None):
        r"""Reduce using SOR-IRKA.

        It uses IRKA as the intermediate reductor, to reduce from 2r to
        r poles. See Section 5.3.2 in :cite:`W12`.

        Parameters
        ----------
        rom0_params
            Can be:

            - order of the reduced model (a positive integer),
            - dict with `'sigma'`, `'b'`, `'c'` as keys mapping to
              initial interpolation points (a 1D |NumPy array|), right
              tangential directions (|VectorArray| from
              `fom.D.source`), and left tangential directions
              (|VectorArray| from `fom.D.range`), all of the same
              length (the order of the reduced model),
            - initial reduced-order model (|LTIModel|).

            If the order of reduced model is given, initial
            interpolation data is generated randomly.
        tol
            Tolerance for the convergence criterion.
        maxit
            Maximum number of iterations.
        num_prev
            Number of previous iterations to compare the current
            iteration to. Larger number can avoid occasional cyclic
            behavior of IRKA.
        force_sigma_in_rhp
            If `False`, new interpolation are reflections of the current
            reduced order model's poles. Otherwise, only the poles in
            the left half-plane are reflected.
        projection
            Projection method:

            - `'orth'`: projection matrices are orthogonalized with
              respect to the Euclidean inner product
            - `'biorth'`: projection matrices are biorthogolized with
              respect to the E product
        conv_crit
            Convergence criterion:

            - `'sigma'`: relative change in interpolation points
            - `'h2'`: relative :math:`\mathcal{H}_2` distance of
              reduced-order models
        compute_errors
            Should the relative :math:`\mathcal{H}_2`-errors of
            intermediate reduced order models be computed.

            .. warning::
                Computing :math:`\mathcal{H}_2`-errors is expensive. Use
                this option only if necessary.
        irka_options
            Dict of options for IRKAReductor.reduce.

        Returns
        -------
        rom
            Reduced-order |SecondOrderModel|.
        """
        if not self.fom.cont_time:
            raise NotImplementedError

        self._clear_lists()
        sigma, b, c = self._rom0_params_to_sigma_b_c(rom0_params, force_sigma_in_rhp)
        self._store_sigma_b_c(sigma, b, c)
        self._check_common_args(tol, maxit, num_prev, conv_crit)
        assert projection in ('orth', 'biorth')
        assert irka_options is None or isinstance(irka_options, dict)
        if not irka_options:
            irka_options = {}

        self.logger.info('Starting SOR-IRKA')
        self._conv_data = (num_prev + 1) * [None]
        if conv_crit == 'sigma':
            self._conv_data[0] = sigma
        self._pg_reductor = SOBHIReductor(self.fom, mu=self.mu)
        for it in range(maxit):
            rom = self._pg_reductor.reduce(sigma, b, c, projection=projection)
            with self.logger.block('Intermediate reduction ...'):
                irka_reductor = IRKAReductor(rom.to_lti())
                rom_r = irka_reductor.reduce(rom.order, **irka_options)
            sigma, b, c = self._rom_to_sigma_b_c(rom_r, force_sigma_in_rhp)
            self._store_sigma_b_c(sigma, b, c)
            self._update_conv_data(sigma, rom, conv_crit)
            self._compute_conv_crit(rom, conv_crit, it)
            self._compute_error(rom, it, compute_errors)
            if self.conv_crit[-1] < tol:
                break

        self.V = self._pg_reductor.V
        self.W = self._pg_reductor.W
        return rom



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from pymor.operators.interface import Operator
from pymor.vectorarrays.list import CopyOnWriteVector, ListVectorSpace

import numpy as np
import math
from model import Vector, DiffusionOperator


class WrappedVector(CopyOnWriteVector):

    def __init__(self, vector):
        assert isinstance(vector, Vector)
        self._impl = vector

    @classmethod
    def from_instance(cls, instance):
        return cls(instance._impl)

    def to_numpy(self, ensure_copy=False):
        result = np.frombuffer(self._impl, dtype=np.float64)
        if ensure_copy:
            result = result.copy()
        return result

    def _copy_data(self):
        self._impl = Vector(self._impl)

    def _scal(self, alpha):
        self._impl.scal(alpha)

    def _axpy(self, alpha, x):
        self._impl.axpy(alpha, x._impl)

    def inner(self, other):
        return self._impl.inner(other._impl)

    def norm(self):
        return math.sqrt(self.inner(self))

    def norm2(self):
        return self.inner(self)

    def sup_norm(self):
        raise NotImplementedError

    def dofs(self, dof_indices):
        raise NotImplementedError

    def amax(self):
        raise NotImplementedError


class WrappedVectorSpace(ListVectorSpace):

    def __init__(self, dim):
        self.dim = dim

    def zero_vector(self):
        return WrappedVector(Vector(self.dim, 0))

    def make_vector(self, obj):
        return WrappedVector(obj)

    def __eq__(self, other):
        return type(other) is WrappedVectorSpace and self.dim == other.dim


class WrappedDiffusionOperator(Operator):
    def __init__(self, op):
        assert isinstance(op, DiffusionOperator)
        self.op = op
        self.source = WrappedVectorSpace(op.dim_source)
        self.range = WrappedVectorSpace(op.dim_range)
        self.linear = True

    @classmethod
    def create(cls, n, left, right):
        return cls(DiffusionOperator(n, left, right))

    def apply(self, U, mu=None):
        assert U in self.source

        def apply_one_vector(u):
            v = Vector(self.range.dim, 0)
            self.op.apply(u._impl, v)
            return v

        return self.range.make_array([apply_one_vector(u) for u in U._list])



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from inspect import getsource

from pymor.core.config import is_jupyter


def format_source(obj):
    """Format source code of an object.

    Parameters
    ----------
    obj
        The object of which to format the source code.

    Returns
    -------
    source
        The source code as a string, possibly highlighted when used in a
        terminal.
    """
    source = getsource(obj)
    if is_jupyter():
        return source
    try:
        from pygments import highlight
        from pygments.lexers import PythonLexer
        from pygments.formatters import Terminal256Formatter
        return highlight(source, PythonLexer(), Terminal256Formatter())
    except ImportError:
        return source


def print_source(obj):
    """Print source code of an object.

    Parameters
    ----------
    obj
        The object of which to print the source code.
    """
    source = format_source(obj)
    if is_jupyter():
        from IPython.display import Code, display
        display(Code(source, language='python'))
    else:
        print(source)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import os
from itertools import chain, product

import numpy as np
import pytest
import scipy.linalg as spla
import scipy.sparse as sps

from pymor.algorithms.lyapunov import solve_lyap_dense, solve_lyap_lrcf
from pymor.core.config import config
from pymor.operators.numpy import NumpyMatrixOperator

n_list_small = [10, 20]
n_list_big = [300]
m_list = [1, 2]
lyap_lrcf_solver_list = [
    'pymess_lradi',
    'lradi',
]
lyap_dense_solver_list = [
    'scipy',
    'slycot_bartels-stewart',
    'pymess_glyap',
]


def fro_norm(A):
    if not sps.issparse(A):
        return spla.norm(A)
    else:
        return sps.linalg.norm(A)


def conv_diff_1d_fd(n, a, b):
    diagonals = [-a * 2 * (n + 1) ** 2 * np.ones((n,)),
                 (a * (n + 1) ** 2 + b * (n + 1) / 2) * np.ones((n - 1,)),
                 (a * (n + 1) ** 2 - b * (n + 1) / 2) * np.ones((n - 1,))]
    A = sps.diags(diagonals, [0, -1, 1], format='csc')
    return A


def conv_diff_1d_fem(n, a, b):
    diagonals = [-a * 2 * (n + 1) ** 2 * np.ones((n,)),
                 (a * (n + 1) ** 2 + b * (n + 1) / 2) * np.ones((n - 1,)),
                 (a * (n + 1) ** 2 - b * (n + 1) / 2) * np.ones((n - 1,))]
    A = sps.diags(diagonals, [0, -1, 1], format='csc')
    diagonals = [2 / 3 * np.ones((n,)),
                 1 / 6 * np.ones((n - 1,)),
                 1 / 6 * np.ones((n - 1,))]
    E = sps.diags(diagonals, [0, -1, 1], format='csc')
    return A, E


def relative_residual(A, E, B, X, trans=False):
    if not trans:
        if E is None:
            AX = A @ X
            BBT = B @ B.T
            res = fro_norm(AX + AX.T + BBT)
            rhs = fro_norm(BBT)
        else:
            AXET = A @ X @ E.T
            BBT = B @ B.T
            res = fro_norm(AXET + AXET.T + BBT)
            rhs = fro_norm(BBT)
    else:
        if E is None:
            ATX = A.T @ X
            CTC = B.T @ B
            res = fro_norm(ATX + ATX.T + CTC)
            rhs = fro_norm(CTC)
        else:
            ATXE = A.T @ X @ E
            CTC = B.T @ B
            res = fro_norm(ATXE + ATXE.T + CTC)
            rhs = fro_norm(CTC)
    return res / rhs


def _check_availability(lyap_solver):
    if (lyap_solver.startswith('slycot')
            and not os.environ.get('DOCKER_PYMOR', False)
            and not config.HAVE_SLYCOT):
        pytest.skip('slycot not available')
    if (lyap_solver.startswith('pymess')
            and not os.environ.get('DOCKER_PYMOR', False)
            and not config.HAVE_PYMESS):
        pytest.skip('pymess not available')


@pytest.mark.parametrize('m', m_list)
@pytest.mark.parametrize('with_E', [False, True])
@pytest.mark.parametrize('trans', [False, True])
@pytest.mark.parametrize('n,lyap_solver', chain(product(n_list_small, lyap_dense_solver_list),
                                                product(n_list_big, lyap_lrcf_solver_list)))
def test_lrcf(n, m, with_E, trans, lyap_solver):
    _check_availability(lyap_solver)

    if not with_E:
        A = conv_diff_1d_fd(n, 1, 0.1)
        E = None
    else:
        A, E = conv_diff_1d_fem(n, 1, 0.1)
    np.random.seed(0)
    B = np.random.randn(n, m)
    if trans:
        B = B.T

    Aop = NumpyMatrixOperator(A)
    Eop = NumpyMatrixOperator(E) if with_E else None
    Bva = Aop.source.from_numpy(B.T if not trans else B)

    Zva = solve_lyap_lrcf(Aop, Eop, Bva, trans=trans, options=lyap_solver)
    assert len(Zva) <= n

    Z = Zva.to_numpy().T
    assert relative_residual(A, E, B, Z @ Z.T, trans=trans) < 1e-10


@pytest.mark.parametrize('n', n_list_small)
@pytest.mark.parametrize('m', m_list)
@pytest.mark.parametrize('with_E', [False, True])
@pytest.mark.parametrize('trans', [False, True])
@pytest.mark.parametrize('lyap_solver', lyap_dense_solver_list)
def test_dense(n, m, with_E, trans, lyap_solver):
    _check_availability(lyap_solver)

    np.random.seed(0)
    A = np.random.randn(n, n)
    E = np.eye(n) + np.random.randn(n, n) / n if with_E else None
    B = np.random.randn(n, m)
    if trans:
        B = B.T

    X = solve_lyap_dense(A, E, B, trans=trans, options=lyap_solver)
    assert type(X) is np.ndarray

    assert relative_residual(A, E, B, X, trans=trans) < 1e-10



#!/usr/bin/env python
# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np
from typer import Argument, run

from pymor.analyticalproblems.domaindescriptions import CircularSectorDomain
from pymor.analyticalproblems.elliptic import StationaryProblem
from pymor.analyticalproblems.functions import ConstantFunction, ExpressionFunction
from pymor.discretizers.builtin import discretize_stationary_cg


def main(
    angle: float = Argument(..., help='The angle of the circular sector.'),
    num_points: int = Argument(..., help='The number of points that form the arc of the circular sector.'),
    clscale: float = Argument(..., help='Mesh element size scaling factor.'),
):
    """Solves the Poisson equation in 2D on a circular sector domain of radius 1
    using an unstructured mesh.

    Note that Gmsh (http://geuz.org/gmsh/) is required for meshing.
    """
    problem = StationaryProblem(
        domain=CircularSectorDomain(angle, radius=1, num_points=num_points),
        diffusion=ConstantFunction(1, dim_domain=2),
        rhs=ConstantFunction(np.array(0.), dim_domain=2, name='rhs'),
        dirichlet_data=ExpressionFunction('sin(angle(x) * pi/rho)', 2,
                                          {}, {'rho': angle}, name='dirichlet')
    )

    print('Discretize ...')
    m, data = discretize_stationary_cg(analytical_problem=problem, diameter=clscale)
    grid = data['grid']
    print(grid)
    print()

    print('Solve ...')
    U = m.solve()

    solution = ExpressionFunction('norm(x)**(pi/rho) * sin(angle(x) * pi/rho)', 2,
                                  {}, {'rho': angle})
    U_ref = U.space.make_array(solution(grid.centers(2)))

    m.visualize((U, U_ref, U-U_ref),
                legend=('Solution', 'Analytical solution (circular boundary)', 'Error'),
                separate_colorbars=True)


if __name__ == '__main__':
    run(main)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from pymor.core.defaults import defaults

import numpy as np


def get_random_state(random_state=None, seed=None):
    """Returns a |NumPy| :class:`~numpy.random.RandomState`.

    Parameters
    ----------
    random_state
        If specified, this state is returned.
    seed
        If specified, the seed to initialize a new random state.

    Returns
    -------
    Either the provided, a newly created or the default `RandomState`
    object.
    """
    assert random_state is None or seed is None
    if random_state is not None:
        return random_state
    elif seed is not None:
        return np.random.RandomState(seed)
    else:
        return default_random_state()


@defaults('seed')
def default_random_state(seed=42):
    """Returns the default |NumPy| :class:`~numpy.random.RandomState`.

    Parameters
    ----------
    seed
        Seed to use for initializing the random state.

    Returns
    -------
    The default `RandomState` object.
    """
    global _default_random_state

    if _default_random_state is None:
        _default_random_state = np.random.RandomState(seed)

    return _default_random_state


_default_random_state = None



#!/usr/bin/env python
# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import sys

import numpy as np
from typer import Argument, Option, run

from pymor.algorithms.greedy import rb_greedy
from pymor.algorithms.ei import interpolate_function
from pymor.algorithms.error import reduction_error_analysis
from pymor.analyticalproblems.domaindescriptions import RectDomain
from pymor.analyticalproblems.elliptic import StationaryProblem
from pymor.analyticalproblems.functions import ConstantFunction, ExpressionFunction
from pymor.discretizers.builtin import discretize_stationary_cg
from pymor.parameters.functionals import ExpressionParameterFunctional
from pymor.reductors.coercive import CoerciveRBReductor


def main(
    ei_snapshots: int = Argument(..., help='Number of snapshots for empirical interpolation.'),
    ei_size: int = Argument(..., help='Number of interpolation DOFs.'),
    snapshots: int = Argument(..., help='Number of snapshots for basis generation.'),
    rb_size: int = Argument(..., help='Size of the reduced basis.'),
    grid: int = Option(100, help='Use grid with 4*NI*NI elements'),
    plot_ei_err: bool = Option(False, help='Plot empirical interpolation error.'),
    plot_solutions: bool = Option(False, help='Plot some example solutions.'),
    test: int = Option(10, help='Number of snapshots to use for stochastic error estimation.'),
):
    """Reduction of a problem without parameter separability using empirical interpolation."""
    problem = StationaryProblem(
        domain=RectDomain(),

        diffusion=ExpressionFunction('(x[1] > 0.5) * x[0]**exponent[0] + 0.1', 2,
                                     parameters={'exponent': 1}),

        rhs=ConstantFunction(1., 2),

        parameter_ranges=(1, 10)
    )

    print('Discretize ...')
    fom, data = discretize_stationary_cg(problem, diameter=1. / grid)

    print(data['grid'])
    print(f'The parameters are {fom.parameters}')

    if plot_solutions:
        print('Showing some solutions')
        Us = ()
        legend = ()
        for mu in problem.parameter_space.sample_uniformly(4):
            print(f"Solving for exponent = {mu['exponent']} ... ")
            sys.stdout.flush()
            Us = Us + (fom.solve(mu),)
            legend = legend + (f"exponent: {mu['exponent']}",)
        fom.visualize(Us, legend=legend, title='Detailed Solutions', block=True)

    diffusion_ei, ei_data = interpolate_function(
        problem.diffusion,
        parameter_sample=problem.parameter_space.sample_uniformly(ei_snapshots),
        evaluation_points=data['grid'].centers(0),
        max_interpolation_dofs=ei_size
    )

    problem_ei = problem.with_(diffusion=diffusion_ei)
    fom_ei, _ = discretize_stationary_cg(problem_ei, diameter=1. / grid)

    if plot_ei_err:
        print('Showing some EI errors')
        ERRs = ()
        legend = ()
        for mu in problem.parameter_space.sample_randomly(2):
            print(f"Solving for exponent = \n{mu['exponent']} ... ")
            sys.stdout.flush()
            U = fom.solve(mu)
            U_EI = fom_ei.solve(mu)
            ERR = U - U_EI
            ERRs = ERRs + (ERR,)
            legend = legend + (f"exponent: {mu['exponent']}",)
            print(f'Error: {np.max(fom.l2_norm(ERR))}')
        fom.visualize(ERRs, legend=legend, title='EI Errors', separate_colorbars=True)

    print('RB generation ...')

    coercivity_estimator = ExpressionParameterFunctional('1.', fom.parameters)
    reductor = CoerciveRBReductor(fom_ei, product=fom.h1_0_semi_product, coercivity_estimator=coercivity_estimator,
                                  check_orthonormality=False)

    greedy_data = rb_greedy(fom_ei, reductor, problem.parameter_space.sample_uniformly(snapshots),
                            max_extensions=rb_size)
    rom = greedy_data['rom']

    results = reduction_error_analysis(rom, fom=fom, reductor=reductor, error_estimator=True,
                                       error_norms=[fom.h1_0_semi_norm], condition=True,
                                       test_mus=problem.parameter_space.sample_randomly(test),
                                       plot=True)
    print(results['summary'])
    import matplotlib.pyplot
    matplotlib.pyplot.show()

    mumax = results['max_error_mus'][0, -1]
    U = fom.solve(mumax)
    U_RB = reductor.reconstruct(rom.solve(mumax))
    fom.visualize((U, U_RB, U - U_RB), legend=('Detailed Solution', 'Reduced Solution', 'Error'),
                  separate_colorbars=True, block=True)


if __name__ == '__main__':
    run(main)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from pathlib import Path

import meshio
from xml.etree.ElementTree import fromstring
from collections import OrderedDict
from xmljson import BadgerFish
from lxml import etree

from pymor.core.config import config
from pymor.core.exceptions import IOLibsMissing


def _read_collection(xml, metadata_key):
    collection = xml['VTKFile']['Collection']
    files = collection['DataSet']
    data = [(fl[f'@{metadata_key}'], _read_single(fl['@file'])) for fl in files]
    data.sort(key=lambda t: t[0])
    return data


def _read_single(filename):
    mesh = meshio.read(filename)
    assert len(mesh.points)
    return mesh


def _get_collection_data(filename):
    path = Path(filename)
    assert path.is_file()
    bf = BadgerFish(dict_type=OrderedDict)
    return path, bf.data(fromstring(open(path, 'rb').read()))


def _get_vtk_type(path):
    """Parse given file until a VTKFile element is found.

    We use the incremental event emitting parser here since we can expect to encounter appended
    binary data in the xml which lxml cannot parse.

    Parameters
    ----------
        path
            vtk file to peek into

    Returns
    -------
    None if no VTKFile element found, else the type attribute of the VTKFile element
    """
    parser = etree.XMLPullParser(events=('start',))
    with open(path, 'rb') as xml:
        for lines in xml.readlines():
            parser.feed(lines)
            for action, element in parser.read_events():
                if element.tag == 'VTKFile':
                    return element.get('type')
    return None


def read_vtkfile(filename, metadata_key='timestep'):
    """Try to read a given file into a Sequence of meshio.Mesh instances

    Parameters
    ----------
    metadata_key
        Which metadata to extract and return alongside the meshio.Mesh instances.

    Returns
    -------
    A list of (metadata_value, meshio.Mesh) tuples. The length of the list is either 1 for
        a singular vtk/vtu/vtp input file (None is returned as metadata),
        or however many members are in the collection file (pvd).
    """
    from pymor.tools.io import change_to_directory
    vtk_type = _get_vtk_type(filename)
    if vtk_type == 'Collection':
        path, xml = _get_collection_data(filename)
        with change_to_directory(path.parent):
            return _read_collection(xml, metadata_key=metadata_key)
    return [(None, _read_single(filename, vtk_type))]


def write_vtk_collection(filename_base, meshes, metadata=None):
    """Output grid-associated data in vtk format

    filename_base
        common component for output files in collection
    meshes
        Sequence of meshio.Mesh objects
    metadata
        dict of {key1: sequence1, key2: sequence2} where sequence must be of len(meshes) or len == 1
        currently supported keys are "timestep", "name", "group" and "part"
        used to describe datapoints in Vtk collection file
        defaults to { 'timestep': list(range(len(meshes))) }

    Returns
    -------
    full filename of saved file
    """
    if not config.HAVE_VTKIO:
        raise IOLibsMissing()
    from pyevtk.vtk import VtkGroup

    fn_tpl = '{}_{:08d}.vtu'
    metadata = metadata or {'timestep': list(range(len(meshes)))}

    def _meta(key, i):
        if key in metadata.keys():
            return metadata[key][0] if len(metadata[key]) == 1 else metadata[key][i]
        # carry over defaults from pyevtk to not break backwards compat
        return {'timestep': 0, 'group': '', 'name': '', 'part': '0'}[key]

    group = VtkGroup(filename_base)
    for i, mesh in enumerate(meshes):
        fn = fn_tpl.format(filename_base, i)
        mesh.write(fn)
        group.addFile(filepath=fn, sim_time=_meta('timestep', i), group=_meta('group', i), name=_meta('name', i),
                      part=_meta('part', i))
    group.save()
    return f'{filename_base}.pvd'



#!/usr/bin/env python
# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import sys
import time

from typer import Argument, Option, run

from pymor.algorithms.error import reduction_error_analysis
from pymor.core.pickle import dump
from pymor.parallel.default import new_parallel_pool
from pymor.tools.typer import Choices


def main(
    xblocks: int = Argument(..., help='Number of blocks in x direction.'),
    yblocks: int = Argument(..., help='Number of blocks in y direction.'),
    snapshots: int = Argument(
        ...,
        help='naive: ignored\n\n'
             'greedy/pod: Number of training_set parameters per block '
             '(in total SNAPSHOTS^(XBLOCKS * YBLOCKS) parameters).\n\n'
             'adaptive_greedy: size of validation set.\n\n'
    ),
    rbsize: int = Argument(..., help='Size of the reduced basis.'),

    adaptive_greedy_gamma: float = Option(0.2, help='See pymor.algorithms.adaptivegreedy.'),
    adaptive_greedy_rho: float = Option(1.1, help='See pymor.algorithms.adaptivegreedy.'),
    adaptive_greedy_theta: float = Option(0., help='See pymor.algorithms.adaptivegreedy.'),
    alg: Choices('naive greedy adaptive_greedy pod') = Option('greedy', help='The model reduction algorithm to use.'),
    cache_region: Choices('none memory disk persistent') = Option(
        'none',
        help='Name of cache region to use for caching solution snapshots.'
    ),
    extension_alg: Choices('trivial gram_schmidt') = Option(
        'gram_schmidt',
        help='Basis extension algorithm to be used.'
    ),
    fenics: bool = Option(False, help='Use FEniCS model.'),
    greedy_with_error_estimator: bool = Option(True, help='Use error estimator for basis generation.'),
    grid: int = Option(100, help='Use grid with 4*NI*NI elements'),
    ipython_engines: int = Option(
        None,
        help='If positive, the number of IPython cluster engines to use for '
             'parallel greedy search. If zero, no parallelization is performed.'
    ),
    ipython_profile: str = Option(None, help='IPython profile to use for parallelization.'),
    list_vector_array: bool = Option(
        False,
        help='Solve using ListVectorArray[NumpyVector] instead of NumpyVectorArray.'
    ),
    order: int = Option(1, help='Polynomial order of the Lagrange finite elements to use in FEniCS.'),
    pickle: str = Option(
        None,
        help='Pickle reduced model, as well as reductor and high-dimensional model '
             'to files with this prefix.'
    ),
    product: Choices('euclidean h1') = Option(
        'h1',
        help='Product w.r.t. which to orthonormalize and calculate Riesz representatives.'
    ),
    plot_err: bool = Option(False, help='Plot error'),
    plot_error_sequence: bool = Option(False, help='Plot reduction error vs. basis size.'),
    plot_solutions: bool = Option(False, help='Plot some example solutions.'),
    reductor: Choices('traditional residual_basis') = Option(
        'residual_basis',
        help='Reductor (error estimator) to choose.'
    ),
    test: int = Option(10, help='Use COUNT snapshots for stochastic error estimation.'),
):
    """Thermalblock demo."""
    if fenics and cache_region != 'none':
        raise ValueError('Caching of high-dimensional solutions is not supported for FEniCS model.')
    if not fenics and order != 1:
        raise ValueError('Higher-order finite elements only supported for FEniCS model.')

    pool = new_parallel_pool(ipython_num_engines=ipython_engines, ipython_profile=ipython_profile)

    if fenics:
        fom, fom_summary = discretize_fenics(xblocks, yblocks, grid, order)
    else:
        fom, fom_summary = discretize_pymor(xblocks, yblocks, grid, list_vector_array)

    parameter_space = fom.parameters.space(0.1, 1.)

    if cache_region != 'none':
        # building a cache_id is only needed for persistent CacheRegions
        cache_id = (f"pymordemos.thermalblock {fenics} {xblocks} {yblocks}"
                    f"{grid} {order}")
        fom.enable_caching(cache_region.value, cache_id)

    if plot_solutions:
        print('Showing some solutions')
        Us = ()
        legend = ()
        for mu in parameter_space.sample_randomly(2):
            print(f"Solving for diffusion = \n{mu['diffusion']} ... ")
            sys.stdout.flush()
            Us = Us + (fom.solve(mu),)
            legend = legend + (str(mu['diffusion']),)
        fom.visualize(Us, legend=legend, title='Detailed Solutions for different parameters',
                      separate_colorbars=False, block=True)

    print('RB generation ...')

    # define estimator for coercivity constant
    from pymor.parameters.functionals import ExpressionParameterFunctional
    coercivity_estimator = ExpressionParameterFunctional('min(diffusion)', fom.parameters)

    # inner product for computation of Riesz representatives
    product = fom.h1_0_semi_product if product == 'h1' else None

    if reductor == 'residual_basis':
        from pymor.reductors.coercive import CoerciveRBReductor
        reductor = CoerciveRBReductor(fom, product=product, coercivity_estimator=coercivity_estimator,
                                      check_orthonormality=False)
    elif reductor == 'traditional':
        from pymor.reductors.coercive import SimpleCoerciveRBReductor
        reductor = SimpleCoerciveRBReductor(fom, product=product, coercivity_estimator=coercivity_estimator,
                                            check_orthonormality=False)
    else:
        assert False  # this should never happen

    if alg == 'naive':
        rom, red_summary = reduce_naive(fom=fom, reductor=reductor, parameter_space=parameter_space,
                                        basis_size=rbsize)
    elif alg == 'greedy':
        parallel = greedy_with_error_estimator or not fenics  # cannot pickle FEniCS model
        rom, red_summary = reduce_greedy(fom=fom, reductor=reductor, parameter_space=parameter_space,
                                         snapshots_per_block=snapshots,
                                         extension_alg_name=extension_alg.value,
                                         max_extensions=rbsize,
                                         use_error_estimator=greedy_with_error_estimator,
                                         pool=pool if parallel else None)
    elif alg == 'adaptive_greedy':
        parallel = greedy_with_error_estimator or not fenics  # cannot pickle FEniCS model
        rom, red_summary = reduce_adaptive_greedy(fom=fom, reductor=reductor, parameter_space=parameter_space,
                                                  validation_mus=snapshots,
                                                  extension_alg_name=extension_alg.value,
                                                  max_extensions=rbsize,
                                                  use_error_estimator=greedy_with_error_estimator,
                                                  rho=adaptive_greedy_rho,
                                                  gamma=adaptive_greedy_gamma,
                                                  theta=adaptive_greedy_theta,
                                                  pool=pool if parallel else None)
    elif alg == 'pod':
        rom, red_summary = reduce_pod(fom=fom, reductor=reductor, parameter_space=parameter_space,
                                      snapshots_per_block=snapshots,
                                      basis_size=rbsize)
    else:
        assert False  # this should never happen

    if pickle:
        print(f"\nWriting reduced model to file {pickle}_reduced ...")
        with open(pickle + '_reduced', 'wb') as f:
            dump((rom, parameter_space), f)
        if not fenics:  # FEniCS data structures do not support serialization
            print(f"Writing detailed model and reductor to file {pickle}_detailed ...")
            with open(pickle + '_detailed', 'wb') as f:
                dump((fom, reductor), f)

    print('\nSearching for maximum error on random snapshots ...')

    results = reduction_error_analysis(rom,
                                       fom=fom,
                                       reductor=reductor,
                                       error_estimator=True,
                                       error_norms=(fom.h1_0_semi_norm, fom.l2_norm),
                                       condition=True,
                                       test_mus=parameter_space.sample_randomly(test, seed=999),
                                       basis_sizes=0 if plot_error_sequence else 1,
                                       plot=plot_error_sequence,
                                       pool=None if fenics else pool  # cannot pickle FEniCS model
                                       )

    print('\n*** RESULTS ***\n')
    print(fom_summary)
    print(red_summary)
    print(results['summary'])
    sys.stdout.flush()

    if plot_error_sequence:
        import matplotlib.pyplot
        matplotlib.pyplot.show()
    if plot_err:
        mumax = results['max_error_mus'][0, -1]
        U = fom.solve(mumax)
        URB = reductor.reconstruct(rom.solve(mumax))
        fom.visualize((U, URB, U - URB), legend=('Detailed Solution', 'Reduced Solution', 'Error'),
                      title='Maximum Error Solution', separate_colorbars=True, block=True)

    global test_results
    test_results = results


def discretize_pymor(xblocks, yblocks, grid_num_intervals, use_list_vector_array):
    from pymor.analyticalproblems.thermalblock import thermal_block_problem
    from pymor.discretizers.builtin import discretize_stationary_cg
    from pymor.discretizers.builtin.list import convert_to_numpy_list_vector_array

    print('Discretize ...')
    # setup analytical problem
    problem = thermal_block_problem(num_blocks=(xblocks, yblocks))

    # discretize using continuous finite elements
    fom, _ = discretize_stationary_cg(problem, diameter=1. / grid_num_intervals)

    if use_list_vector_array:
        fom = convert_to_numpy_list_vector_array(fom)

    summary = f'''pyMOR model:
   number of blocks: {xblocks}x{yblocks}
   grid intervals:   {grid_num_intervals}
   ListVectorArray:  {use_list_vector_array}
'''

    return fom, summary


def discretize_fenics(xblocks, yblocks, grid_num_intervals, element_order):
    from pymor.tools import mpi

    if mpi.parallel:
        from pymor.models.mpi import mpi_wrap_model
        fom = mpi_wrap_model(lambda: _discretize_fenics(xblocks, yblocks, grid_num_intervals, element_order),
                             use_with=True, pickle_local_spaces=False)
    else:
        fom = _discretize_fenics(xblocks, yblocks, grid_num_intervals, element_order)

    summary = f'''FEniCS model:
   number of blocks:      {xblocks}x{yblocks}
   grid intervals:        {grid_num_intervals}
   finite element order:  {element_order}
'''

    return fom, summary


def _discretize_fenics(xblocks, yblocks, grid_num_intervals, element_order):

    # assemble system matrices - FEniCS code
    ########################################

    import dolfin as df
    mesh = df.UnitSquareMesh(grid_num_intervals, grid_num_intervals, 'crossed')
    V = df.FunctionSpace(mesh, 'Lagrange', element_order)
    u = df.TrialFunction(V)
    v = df.TestFunction(V)

    diffusion = df.Expression('(lower0 <= x[0]) * (open0 ? (x[0] < upper0) : (x[0] <= upper0)) *'
                              '(lower1 <= x[1]) * (open1 ? (x[1] < upper1) : (x[1] <= upper1))',
                              lower0=0., upper0=0., open0=0,
                              lower1=0., upper1=0., open1=0,
                              element=df.FunctionSpace(mesh, 'DG', 0).ufl_element())

    def assemble_matrix(x, y, nx, ny):
        diffusion.user_parameters['lower0'] = x/nx
        diffusion.user_parameters['lower1'] = y/ny
        diffusion.user_parameters['upper0'] = (x + 1)/nx
        diffusion.user_parameters['upper1'] = (y + 1)/ny
        diffusion.user_parameters['open0'] = (x + 1 == nx)
        diffusion.user_parameters['open1'] = (y + 1 == ny)
        return df.assemble(df.inner(diffusion * df.nabla_grad(u), df.nabla_grad(v)) * df.dx)

    mats = [assemble_matrix(x, y, xblocks, yblocks)
            for x in range(xblocks) for y in range(yblocks)]
    mat0 = mats[0].copy()
    mat0.zero()
    h1_mat = df.assemble(df.inner(df.nabla_grad(u), df.nabla_grad(v)) * df.dx)
    l2_mat = df.assemble(u * v * df.dx)

    f = df.Constant(1.) * v * df.dx
    F = df.assemble(f)

    bc = df.DirichletBC(V, 0., df.DomainBoundary())
    for m in mats:
        bc.zero(m)
    bc.apply(mat0)
    bc.apply(h1_mat)
    bc.apply(F)

    # wrap everything as a pyMOR model
    ##################################

    # FEniCS wrappers
    from pymor.bindings.fenics import FenicsVectorSpace, FenicsMatrixOperator, FenicsVisualizer

    # generic pyMOR classes
    from pymor.models.basic import StationaryModel
    from pymor.operators.constructions import LincombOperator, VectorOperator
    from pymor.parameters.functionals import ProjectionParameterFunctional

    # define parameter functionals (same as in pymor.analyticalproblems.thermalblock)
    def parameter_functional_factory(x, y):
        return ProjectionParameterFunctional('diffusion',
                                             size=yblocks*xblocks,
                                             index=yblocks - y - 1 + x * yblocks,
                                             name=f'diffusion_{x}_{y}')
    parameter_functionals = tuple(parameter_functional_factory(x, y)
                                  for x in range(xblocks) for y in range(yblocks))

    # wrap operators
    ops = [FenicsMatrixOperator(mat0, V, V)] + [FenicsMatrixOperator(m, V, V) for m in mats]
    op = LincombOperator(ops, (1.,) + parameter_functionals)
    rhs = VectorOperator(FenicsVectorSpace(V).make_array([F]))
    h1_product = FenicsMatrixOperator(h1_mat, V, V, name='h1_0_semi')
    l2_product = FenicsMatrixOperator(l2_mat, V, V, name='l2')

    # build model
    visualizer = FenicsVisualizer(FenicsVectorSpace(V))
    fom = StationaryModel(op, rhs, products={'h1_0_semi': h1_product,
                                             'l2': l2_product},
                          visualizer=visualizer)

    return fom


def reduce_naive(fom, reductor, parameter_space, basis_size):

    tic = time.perf_counter()

    training_set = parameter_space.sample_randomly(basis_size)

    for mu in training_set:
        reductor.extend_basis(fom.solve(mu), method='trivial')

    rom = reductor.reduce()

    elapsed_time = time.perf_counter() - tic

    summary = f'''Naive basis generation:
   basis size set: {basis_size}
   elapsed time:   {elapsed_time}
'''

    return rom, summary


def reduce_greedy(fom, reductor, parameter_space, snapshots_per_block,
                  extension_alg_name, max_extensions, use_error_estimator, pool):

    from pymor.algorithms.greedy import rb_greedy

    # run greedy
    training_set = parameter_space.sample_uniformly(snapshots_per_block)
    greedy_data = rb_greedy(fom, reductor, training_set,
                            use_error_estimator=use_error_estimator, error_norm=fom.h1_0_semi_norm,
                            extension_params={'method': extension_alg_name}, max_extensions=max_extensions,
                            pool=pool)
    rom = greedy_data['rom']

    # generate summary
    real_rb_size = rom.solution_space.dim
    training_set_size = len(training_set)
    summary = f'''Greedy basis generation:
   size of training set:   {training_set_size}
   error estimator used:   {use_error_estimator}
   extension method:       {extension_alg_name}
   prescribed basis size:  {max_extensions}
   actual basis size:      {real_rb_size}
   elapsed time:           {greedy_data["time"]}
'''

    return rom, summary


def reduce_adaptive_greedy(fom, reductor, parameter_space, validation_mus,
                           extension_alg_name, max_extensions, use_error_estimator,
                           rho, gamma, theta, pool):

    from pymor.algorithms.adaptivegreedy import rb_adaptive_greedy

    # run greedy
    greedy_data = rb_adaptive_greedy(fom, reductor, parameter_space, validation_mus=-validation_mus,
                                     use_error_estimator=use_error_estimator, error_norm=fom.h1_0_semi_norm,
                                     extension_params={'method': extension_alg_name}, max_extensions=max_extensions,
                                     rho=rho, gamma=gamma, theta=theta, pool=pool)
    rom = greedy_data['rom']

    # generate summary
    real_rb_size = rom.solution_space.dim
    # the validation set consists of `validation_mus` random parameters plus the centers of the
    # adaptive sample set cells
    validation_mus += 1
    summary = f'''Adaptive greedy basis generation:
   initial size of validation set:  {validation_mus}
   error estimator used:            {use_error_estimator}
   extension method:                {extension_alg_name}
   prescribed basis size:           {max_extensions}
   actual basis size:               {real_rb_size}
   elapsed time:                    {greedy_data["time"]}
'''

    return rom, summary


def reduce_pod(fom, reductor, parameter_space, snapshots_per_block, basis_size):
    from pymor.algorithms.pod import pod

    tic = time.perf_counter()

    training_set = parameter_space.sample_uniformly(snapshots_per_block)

    print('Solving on training set ...')
    snapshots = fom.operator.source.empty(reserve=len(training_set))
    for mu in training_set:
        snapshots.append(fom.solve(mu))

    print('Performing POD ...')
    basis, singular_values = pod(snapshots, modes=basis_size, product=reductor.products['RB'])

    print('Reducing ...')
    reductor.extend_basis(basis, method='trivial')
    rom = reductor.reduce()

    elapsed_time = time.perf_counter() - tic

    # generate summary
    real_rb_size = rom.solution_space.dim
    training_set_size = len(training_set)
    summary = f'''POD basis generation:
   size of training set:   {training_set_size}
   prescribed basis size:  {basis_size}
   actual basis size:      {real_rb_size}
   elapsed time:           {elapsed_time}
'''

    return rom, summary


if __name__ == '__main__':
    run(main)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""Remark on the documentation:

Due to an issue in autoapi, the classes `NeuralNetworkStatefreeOutputReductor`,
`NeuralNetworkInstationaryReductor`, `NeuralNetworkInstationaryStatefreeOutputReductor`,
`EarlyStoppingScheduler` and `CustomDataset` do not appear in the documentation,
see https://github.com/pymor/pymor/issues/1343.
"""

from pymor.core.config import config


if config.HAVE_TORCH:
    from numbers import Number

    import numpy as np

    import torch
    import torch.nn as nn
    import torch.optim as optim
    import torch.utils as utils

    from pymor.algorithms.pod import pod
    from pymor.algorithms.projection import project
    from pymor.core.base import BasicObject
    from pymor.core.exceptions import NeuralNetworkTrainingFailed
    from pymor.core.logger import getLogger
    from pymor.models.neural_network import (FullyConnectedNN, NeuralNetworkModel,
                                             NeuralNetworkStatefreeOutputModel,
                                             NeuralNetworkInstationaryModel,
                                             NeuralNetworkInstationaryStatefreeOutputModel)

    class NeuralNetworkReductor(BasicObject):
        """Reduced Basis reductor relying on artificial neural networks.

        This is a reductor that constructs a reduced basis using proper
        orthogonal decomposition and trains a neural network that approximates
        the mapping from parameter space to coefficients of the full-order
        solution in the reduced basis.
        The approach is described in :cite:`HU18`.

        Parameters
        ----------
        fom
            The full-order |Model| to reduce.
        training_set
            Set of |parameter values| to use for POD and training of the
            neural network.
        validation_set
            Set of |parameter values| to use for validation in the training
            of the neural network.
        validation_ratio
            Fraction of the training set to use for validation in the training
            of the neural network (only used if no validation set is provided).
        basis_size
            Desired size of the reduced basis. If `None`, rtol, atol or l2_err must
            be provided.
        rtol
            Relative tolerance the basis should guarantee on the training set.
        atol
            Absolute tolerance the basis should guarantee on the training set.
        l2_err
            L2-approximation error the basis should not exceed on the training
            set.
        pod_params
            Dict of additional parameters for the POD-method.
        ann_mse
            If `'like_basis'`, the mean squared error of the neural network on
            the training set should not exceed the error of projecting onto the basis.
            If `None`, the neural network with smallest validation error is
            used to build the ROM.
            If a tolerance is prescribed, the mean squared error of the neural
            network on the training set should not exceed this threshold.
            Training is interrupted if a neural network that undercuts the
            error tolerance is found.
        """

        def __init__(self, fom, training_set, validation_set=None, validation_ratio=0.1,
                     basis_size=None, rtol=0., atol=0., l2_err=0., pod_params=None,
                     ann_mse='like_basis'):
            assert 0 < validation_ratio < 1 or validation_set
            self.__auto_init(locals())

        def reduce(self, hidden_layers='[(N+P)*3, (N+P)*3]', activation_function=torch.tanh,
                   optimizer=optim.LBFGS, epochs=1000, batch_size=20, learning_rate=1.,
                   restarts=10, seed=0):
            """Reduce by training artificial neural networks.

            Parameters
            ----------
            hidden_layers
                Number of neurons in the hidden layers. Can either be fixed or
                a Python expression string depending on the reduced basis size
                respectively output dimension `N` and the total dimension of
                the |Parameters| `P`.
            activation_function
                Activation function to use between the hidden layers.
            optimizer
                Algorithm to use as optimizer during training.
            epochs
                Maximum number of epochs for training.
            batch_size
                Batch size to use if optimizer allows mini-batching.
            learning_rate
                Step size to use in each optimization step.
            restarts
                Number of restarts of the training algorithm. Since the training
                results highly depend on the initial starting point, i.e. the
                initial weights and biases, it is advisable to train multiple
                neural networks by starting with different initial values and
                choose that one performing best on the validation set.
            seed
                Seed to use for various functions in PyTorch. Using a fixed seed,
                it is possible to reproduce former results.

            Returns
            -------
            rom
                Reduced-order |NeuralNetworkModel|.
            """
            assert restarts > 0
            assert epochs > 0
            assert batch_size > 0
            assert learning_rate > 0.

            # set a seed for the PyTorch initialization of weights and biases
            # and further PyTorch methods
            torch.manual_seed(seed)

            # build a reduced basis using POD and compute training data
            if not hasattr(self, 'training_data'):
                self.compute_training_data()

            layer_sizes = self._compute_layer_sizes(hidden_layers)

            # compute validation data
            if not hasattr(self, 'validation_data'):
                with self.logger.block('Computing validation snapshots ...'):

                    if self.validation_set:
                        self.validation_data = []
                        for mu in self.validation_set:
                            sample = self._compute_sample(mu)
                            self.validation_data.extend(sample)
                    else:
                        number_validation_snapshots = int(len(self.training_data)*self.validation_ratio)
                        # randomly shuffle training data before splitting into two sets
                        np.random.shuffle(self.training_data)
                        # split training data into validation and training set
                        self.validation_data = self.training_data[0:number_validation_snapshots]
                        self.training_data = self.training_data[number_validation_snapshots+1:]

            # run the actual training of the neural network
            with self.logger.block('Training of neural network ...'):
                target_loss = self._compute_target_loss()
                # set parameters for neural network and training
                neural_network_parameters = {'layer_sizes': layer_sizes,
                                             'activation_function': activation_function}
                training_parameters = {'optimizer': optimizer, 'epochs': epochs,
                                       'batch_size': batch_size, 'learning_rate': learning_rate}

                self.logger.info('Initializing neural network ...')
                # initialize the neural network
                neural_network = FullyConnectedNN(**neural_network_parameters).double()
                # run training algorithm with multiple restarts
                self.neural_network, self.losses = multiple_restarts_training(self.training_data, self.validation_data,
                                                                              neural_network, target_loss, restarts,
                                                                              training_parameters, seed)

            self._check_tolerances()

            return self._build_rom()

        def compute_training_data(self):
            """Compute a reduced basis using proper orthogonal decomposition."""
            # compute snapshots for POD and training of neural networks
            with self.logger.block('Computing training snapshots ...'):
                U = self.fom.solution_space.empty()
                for mu in self.training_set:
                    U.append(self.fom.solve(mu))

            # compute reduced basis via POD
            with self.logger.block('Building reduced basis ...'):
                self.reduced_basis, svals = pod(U, modes=self.basis_size, rtol=self.rtol / 2.,
                                                atol=self.atol / 2., l2_err=self.l2_err / 2.,
                                                **(self.pod_params or {}))

            # compute training samples
            with self.logger.block('Computing training samples ...'):
                self.training_data = []
                for mu, u in zip(self.training_set, U):
                    sample = self._compute_sample(mu, u)
                    self.training_data.extend(sample)

            # compute mean square loss
            self.mse_basis = (sum(U.norm2()) - sum(svals**2)) / len(U)

        def _compute_sample(self, mu, u=None):
            """Transform parameter and corresponding solution to |NumPy arrays|."""
            # determine the coefficients of the full-order solutions in the reduced basis to obtain
            # the training data
            if u is None:
                u = self.fom.solve(mu)
            return [(mu, self.reduced_basis.inner(u)[:, 0])]

        def _compute_layer_sizes(self, hidden_layers):
            """Compute the number of neurons in the layers of the neural network."""
            # determine the numbers of neurons in the hidden layers
            if isinstance(hidden_layers, str):
                hidden_layers = eval(hidden_layers, {'N': len(self.reduced_basis), 'P': self.fom.parameters.dim})
            # input and output size of the neural network are prescribed by the
            # dimension of the parameter space and the reduced basis size
            assert isinstance(hidden_layers, list)
            return [self.fom.parameters.dim, ] + hidden_layers + [len(self.reduced_basis), ]

        def _compute_target_loss(self):
            """Compute target loss depending on value of `ann_mse`."""
            target_loss = None
            if isinstance(self.ann_mse, Number):
                target_loss = self.ann_mse
            elif self.ann_mse == 'like_basis':
                target_loss = self.mse_basis
            return target_loss

        def _check_tolerances(self):
            """Check if trained neural network is sufficient to guarantee certain error bounds."""
            with self.logger.block('Checking tolerances for error of neural network ...'):

                if isinstance(self.ann_mse, Number):
                    if self.losses['full'] > self.ann_mse:
                        raise NeuralNetworkTrainingFailed('Could not train a neural network that '
                                                          'guarantees prescribed tolerance!')
                elif self.ann_mse == 'like_basis':
                    if self.losses['full'] > self.mse_basis:
                        raise NeuralNetworkTrainingFailed('Could not train a neural network with an error as small as '
                                                          'the reduced basis error! Maybe you can try a different '
                                                          'neural network architecture or change the value of '
                                                          '`ann_mse`.')
                elif self.ann_mse is None:
                    self.logger.info('Using neural network with smallest validation error ...')
                    self.logger.info(f'Finished training with a validation loss of {self.losses["val"]} ...')
                else:
                    raise ValueError('Unknown value for mean squared error of neural network')

        def _build_rom(self):
            """Construct the reduced order model."""
            with self.logger.block('Building ROM ...'):
                projected_output_functional = (project(self.fom.output_functional, None, self.reduced_basis)
                                               if self.fom.output_functional else None)
                rom = NeuralNetworkModel(self.neural_network, parameters=self.fom.parameters,
                                         output_functional=projected_output_functional,
                                         name=f'{self.fom.name}_reduced')

            return rom

        def reconstruct(self, u):
            """Reconstruct high-dimensional vector from reduced vector `u`."""
            assert hasattr(self, 'reduced_basis')
            return self.reduced_basis.lincomb(u.to_numpy())

    class NeuralNetworkStatefreeOutputReductor(NeuralNetworkReductor):
        """Output reductor relying on artificial neural networks.

        This is a reductor that trains a neural network that approximates
        the mapping from parameter space to output space.

        Parameters
        ----------
        fom
            The full-order |Model| to reduce.
        training_set
            Set of |parameter values| to use for POD and training of the
            neural network.
        validation_set
            Set of |parameter values| to use for validation in the training
            of the neural network.
        validation_ratio
            Fraction of the training set to use for validation in the training
            of the neural network (only used if no validation set is provided).
        validation_loss
            The validation loss to reach during training. If `None`, the neural
            network with the smallest validation loss is returned.
        """

        def __init__(self, fom, training_set, validation_set=None, validation_ratio=0.1,
                     validation_loss=None):
            assert 0 < validation_ratio < 1 or validation_set
            self.__auto_init(locals())

        def compute_training_data(self):
            """Compute the training samples (the outputs to the parameters of the training set)."""
            with self.logger.block('Computing training samples ...'):
                self.training_data = []
                for mu in self.training_set:
                    sample = self._compute_sample(mu)
                    self.training_data.extend(sample)

        def _compute_sample(self, mu):
            """Transform parameter and corresponding output to tensors."""
            return [(mu, self.fom.output(mu).flatten())]

        def _compute_layer_sizes(self, hidden_layers):
            """Compute the number of neurons in the layers of the neural network."""
            # determine the numbers of neurons in the hidden layers
            if isinstance(hidden_layers, str):
                hidden_layers = eval(hidden_layers, {'N': self.fom.dim_output, 'P': self.fom.parameters.dim})
            # input and output size of the neural network are prescribed by the
            # dimension of the parameter space and the output dimension
            assert isinstance(hidden_layers, list)
            return [self.fom.parameters.dim, ] + hidden_layers + [self.fom.dim_output, ]

        def _compute_target_loss(self):
            """Compute target loss depending on value of `ann_mse`."""
            return self.validation_loss

        def _check_tolerances(self):
            """Check if trained neural network is sufficient to guarantee certain error bounds."""
            self.logger.info('Using neural network with smallest validation error ...')
            self.logger.info(f'Finished training with a validation loss of {self.losses["val"]} ...')

        def _build_rom(self):
            """Construct the reduced order model."""
            with self.logger.block('Building ROM ...'):
                rom = NeuralNetworkStatefreeOutputModel(self.neural_network, self.fom.parameters,
                                                        name=f'{self.fom.name}_output_reduced')

            return rom

    class NeuralNetworkInstationaryReductor(NeuralNetworkReductor):
        """Reduced Basis reductor for instationary problems relying on artificial neural networks.

        This is a reductor that constructs a reduced basis using proper
        orthogonal decomposition and trains a neural network that approximates
        the mapping from parameter and time space to coefficients of the
        full-order solution in the reduced basis.
        The approach is described in :cite:`WHR19`.

        Parameters
        ----------
        fom
            The full-order |Model| to reduce.
        training_set
            Set of |parameter values| to use for POD and training of the
            neural network.
        validation_set
            Set of |parameter values| to use for validation in the training
            of the neural network.
        validation_ratio
            Fraction of the training set to use for validation in the training
            of the neural network (only used if no validation set is provided).
        basis_size
            Desired size of the reduced basis. If `None`, rtol, atol or l2_err must
            be provided.
        rtol
            Relative tolerance the basis should guarantee on the training set.
        atol
            Absolute tolerance the basis should guarantee on the training set.
        l2_err
            L2-approximation error the basis should not exceed on the training
            set.
        pod_params
            Dict of additional parameters for the POD-method.
        ann_mse
            If `'like_basis'`, the mean squared error of the neural network on
            the training set should not exceed the error of projecting onto the basis.
            If `None`, the neural network with smallest validation error is
            used to build the ROM.
            If a tolerance is prescribed, the mean squared error of the neural
            network on the training set should not exceed this threshold.
            Training is interrupted if a neural network that undercuts the
            error tolerance is found.
        """

        def __init__(self, fom, training_set, validation_set=None, validation_ratio=0.1,
                     basis_size=None, rtol=0., atol=0., l2_err=0., pod_params=None,
                     ann_mse='like_basis'):
            assert 0 < validation_ratio < 1 or validation_set
            self.__auto_init(locals())

        def compute_training_data(self):
            """Compute a reduced basis using proper orthogonal decomposition."""
            # compute snapshots for POD and training of neural networks
            with self.logger.block('Computing training snapshots ...'):
                U = self.fom.solution_space.empty()
                for mu in self.training_set:
                    u = self.fom.solve(mu)
                    if hasattr(self, 'nt'):
                        assert self.nt == len(u)
                    else:
                        self.nt = len(u)
                    U.append(u)

            # compute reduced basis via POD
            with self.logger.block('Building reduced basis ...'):
                self.reduced_basis, svals = pod(U, modes=self.basis_size, rtol=self.rtol / 2.,
                                                atol=self.atol / 2., l2_err=self.l2_err / 2.,
                                                **(self.pod_params or {}))

            # compute training samples
            with self.logger.block('Computing training samples ...'):
                self.training_data = []
                for i, mu in enumerate(self.training_set):
                    sample = self._compute_sample(mu, U[i*self.nt:(i+1)*self.nt])
                    self.training_data.extend(sample)

            # compute mean square loss
            self.mse_basis = (sum(U.norm2()) - sum(svals**2)) / len(U)

        def _compute_sample(self, mu, u=None):
            """Transform parameter and corresponding solution to |NumPy arrays|.

            This function takes care of including the time instances in the inputs.
            """
            if u is None:
                u = self.fom.solve(mu)

            parameters_with_time = [mu.with_(t=t) for t in np.linspace(0, self.fom.T, self.nt)]

            samples = [(mu, self.reduced_basis.inner(u_t)[:, 0])
                       for mu, u_t in zip(parameters_with_time, u)]

            return samples

        def _compute_layer_sizes(self, hidden_layers):
            """Compute the number of neurons in the layers of the neural network
            (make sure to increase the input dimension to account for the time).
            """
            # determine the numbers of neurons in the hidden layers
            if isinstance(hidden_layers, str):
                hidden_layers = eval(hidden_layers, {'N': len(self.reduced_basis), 'P': self.fom.parameters.dim})
            # input and output size of the neural network are prescribed by the
            # dimension of the parameter space and the reduced basis size
            assert isinstance(hidden_layers, list)
            return [self.fom.parameters.dim + 1, ] + hidden_layers + [len(self.reduced_basis), ]

        def _build_rom(self):
            """Construct the reduced order model."""
            with self.logger.block('Building ROM ...'):
                projected_output_functional = (project(self.fom.output_functional, None, self.reduced_basis)
                                               if self.fom.output_functional else None)
                rom = NeuralNetworkInstationaryModel(self.fom.T, self.nt, self.neural_network,
                                                     parameters=self.fom.parameters,
                                                     output_functional=projected_output_functional,
                                                     name=f'{self.fom.name}_reduced')

            return rom

    class NeuralNetworkInstationaryStatefreeOutputReductor(NeuralNetworkStatefreeOutputReductor):
        """Output reductor relying on artificial neural networks.

        This is a reductor that trains a neural network that approximates
        the mapping from parameter space to output space.

        Parameters
        ----------
        fom
            The full-order |Model| to reduce.
        nt
            Number of time steps in the reduced order model (does not have to
            coincide with the number of time steps in the full order model).
        training_set
            Set of |parameter values| to use for POD and training of the
            neural network.
        validation_set
            Set of |parameter values| to use for validation in the training
            of the neural network.
        validation_ratio
            Fraction of the training set to use for validation in the training
            of the neural network (only used if no validation set is provided).
        validation_loss
            The validation loss to reach during training. If `None`, the neural
            network with the smallest validation loss is returned.
        """

        def __init__(self, fom, nt, training_set, validation_set=None, validation_ratio=0.1,
                     validation_loss=None):
            assert 0 < validation_ratio < 1 or validation_set
            self.__auto_init(locals())

        def _compute_sample(self, mu):
            """Transform parameter and corresponding output to |NumPy arrays|.

            This function takes care of including the time instances in the inputs.
            """
            output_trajectory = self.fom.output(mu)
            output_size = output_trajectory.shape[0]
            samples = [(mu.with_(t=t), output.flatten())
                       for t, output in zip(np.linspace(0, self.fom.T, output_size), output_trajectory)]

            return samples

        def _compute_layer_sizes(self, hidden_layers):
            """Compute the number of neurons in the layers of the neural network."""
            # determine the numbers of neurons in the hidden layers
            if isinstance(hidden_layers, str):
                hidden_layers = eval(hidden_layers, {'N': self.fom.dim_output, 'P': self.fom.parameters.dim})
            # input and output size of the neural network are prescribed by the
            # dimension of the parameter space and the output dimension
            assert isinstance(hidden_layers, list)
            return [self.fom.parameters.dim + 1, ] + hidden_layers + [self.fom.dim_output, ]

        def _build_rom(self):
            """Construct the reduced order model."""
            with self.logger.block('Building ROM ...'):
                rom = NeuralNetworkInstationaryStatefreeOutputModel(self.fom.T, self.nt, self.neural_network,
                                                                    self.fom.parameters,
                                                                    name=f'{self.fom.name}_output_reduced')

            return rom

    class EarlyStoppingScheduler(BasicObject):
        """Class for performing early stopping in training of neural networks.

        If the validation loss does not decrease over a certain amount of epochs, the
        training should be aborted to avoid overfitting the training data.
        This class implements an early stopping scheduler that recommends to stop the
        training process if the validation loss did not decrease by at least `delta`
        over `patience` epochs.

        Parameters
        ----------
        size_training_validation_set
            Size of both, training and validation set together.
        patience
            Number of epochs of non-decreasing validation loss allowed, before early
            stopping the training process.
        delta
            Minimal amount of decrease in the validation loss that is required to reset
            the counter of non-decreasing epochs.
        """

        def __init__(self, size_training_validation_set, patience=10, delta=0.):
            self.__auto_init(locals())

            self.best_losses = None
            self.best_neural_network = None
            self.counter = 0

        def __call__(self, losses, neural_network=None):
            """Returns `True` if early stopping of training is suggested.

            Parameters
            ----------
            losses
                Dictionary of losses on the validation and the training set in
                the current epoch.
            neural_network
                Neural network that produces the current validation loss.

            Returns
            -------
            `True` if early stopping is suggested, `False` otherwise.
            """
            if self.best_losses is None:
                self.best_losses = losses
                self.best_losses['full'] /= self.size_training_validation_set
                self.best_neural_network = neural_network
            elif self.best_losses['val'] - self.delta <= losses['val']:
                self.counter += 1
                if self.counter >= self.patience:
                    return True
            else:
                self.best_losses = losses
                self.best_losses['full'] /= self.size_training_validation_set
                self.best_neural_network = neural_network
                self.counter = 0

            return False

    class CustomDataset(utils.data.Dataset):
        """Class that represents the dataset to use in PyTorch.

        Parameters
        ----------
        training_data
            Set of training parameters and the respective coefficients of the
            solution in the reduced basis.
        """

        def __init__(self, training_data):
            self.training_data = training_data

        def __len__(self):
            return len(self.training_data)

        def __getitem__(self, idx):
            t = self.training_data[idx]
            return t

    def train_neural_network(training_data, validation_data, neural_network,
                             training_parameters={}):
        """Training algorithm for artificial neural networks.

        Trains a single neural network using the given training and validation data.

        Parameters
        ----------
        training_data
            Data to use during the training phase. Has to be a list of tuples,
            where each tuple consists of two elements that are either
            PyTorch-tensors (`torch.DoubleTensor`) or |NumPy arrays| or pyMOR data
            structures that have `to_numpy()` implemented.
            The first element contains the input data, the second element contains
            the target values.
        validation_data
            Data to use during the validation phase. Has to be a list of tuples,
            where each tuple consists of two elements that are either
            PyTorch-tensors (`torch.DoubleTensor`) or |NumPy arrays| or pyMOR data
            structures that have `to_numpy()` implemented.
            The first element contains the input data, the second element contains
            the target values.
        neural_network
            The neural network to train (can also be a pre-trained model).
            Has to be a PyTorch-Module.
        training_parameters
            Dictionary with additional parameters for the training routine like
            the type of the optimizer, the (maximum) number of epochs, the batch
            size, the learning rate or the loss function to use.
            Possible keys are `'optimizer'` (an optimizer from the PyTorch `optim`
            package; if not provided, the LBFGS-optimizer is taken as default),
            `'epochs'` (an integer that determines the number of epochs to use
            for training the neural network (if training is not interrupted
            prematurely due to early stopping); if not provided, 1000 is taken as
            default value), `'batch_size'` (an integer that determines the number
            of samples to pass to the optimizer at once; if not provided, 20 is
            taken as default value; not used in the case of the LBFGS-optimizer
            since LBFGS does not support mini-batching), `'learning_rate'` (a
            positive real number used as the (initial) step size of the optimizer;
            if not provided, 1 is taken as default value; thus far, no learning
            rate schedulers are supported in this implementation), and
            `'loss_function'` (a loss function from PyTorch; if not provided, the
            MSE loss is taken as default).

        Returns
        -------
        best_neural_network
            The best trained neural network with respect to validation loss.
        losses
            The corresponding losses as a dictionary with keys `'full'` (for the
            full loss containing the training and the validation average loss),
            `'train'` (for the average loss on the training set), and `'val'`
            (for the average loss on the validation set).
        """
        assert isinstance(neural_network, nn.Module)

        for data in training_data, validation_data:
            assert isinstance(data, list)
            assert all(isinstance(datum, tuple) and len(datum) == 2 for datum in data)

        def prepare_datum(datum):
            if not (isinstance(datum, torch.DoubleTensor) or isinstance(datum, np.ndarray)):
                return datum.to_numpy()
            return datum

        training_data = [(prepare_datum(datum[0]), prepare_datum(datum[1])) for datum in training_data]
        validation_data = [(prepare_datum(datum[0]), prepare_datum(datum[1])) for datum in validation_data]

        optimizer = optim.LBFGS if 'optimizer' not in training_parameters else training_parameters['optimizer']
        epochs = 1000 if 'epochs' not in training_parameters else training_parameters['epochs']
        assert isinstance(epochs, int) and epochs > 0
        batch_size = 20 if 'batch_size' not in training_parameters else training_parameters['batch_size']
        assert isinstance(batch_size, int) and batch_size > 0
        learning_rate = 1. if 'learning_rate' not in training_parameters else training_parameters['learning_rate']
        assert learning_rate > 0.
        loss_function = (nn.MSELoss() if 'loss_function' not in training_parameters
                         else training_parameters['loss_function'])

        logger = getLogger('pymor.algorithms.neural_network.train_neural_network')

        # LBFGS-optimizer does not support mini-batching, so the batch size needs to be adjusted
        if optimizer == optim.LBFGS:
            batch_size = max(len(training_data), len(validation_data))

        # initialize optimizer and early stopping scheduler
        optimizer = optimizer(neural_network.parameters(), lr=learning_rate)
        early_stopping_scheduler = EarlyStoppingScheduler(len(training_data) + len(validation_data))

        # create the training and validation sets as well as the respective data loaders
        training_dataset = CustomDataset(training_data)
        validation_dataset = CustomDataset(validation_data)
        training_loader = utils.data.DataLoader(training_dataset, batch_size=batch_size)
        validation_loader = utils.data.DataLoader(validation_dataset, batch_size=batch_size)
        dataloaders = {'train':  training_loader, 'val': validation_loader}

        phases = ['train', 'val']

        logger.info('Starting optimization procedure ...')

        # perform optimization procedure
        for epoch in range(epochs):
            losses = {'full': 0.}

            # alternate between training and validation phase
            for phase in phases:
                if phase == 'train':
                    neural_network.train()
                else:
                    neural_network.eval()

                running_loss = 0.0

                # iterate over batches
                for batch in dataloaders[phase]:
                    inputs = batch[0]
                    targets = batch[1]

                    with torch.set_grad_enabled(phase == 'train'):
                        def closure():
                            if torch.is_grad_enabled():
                                optimizer.zero_grad()
                            outputs = neural_network(inputs)
                            loss = loss_function(outputs, targets)
                            if loss.requires_grad:
                                loss.backward()
                            return loss

                        # perform optimization step
                        if phase == 'train':
                            optimizer.step(closure)

                        # compute loss of current batch
                        loss = closure()

                    # update overall absolute loss
                    running_loss += loss.item() * len(batch[0])

                # compute average loss
                epoch_loss = running_loss / len(dataloaders[phase].dataset)

                losses[phase] = epoch_loss

                losses['full'] += running_loss

                # check for early stopping
                if phase == 'val' and early_stopping_scheduler(losses, neural_network):
                    logger.info(f'Stopping training process early after {epoch + 1} epochs with validation loss '
                                f'of {early_stopping_scheduler.best_losses["val"]:.3e} ...')
                    return early_stopping_scheduler.best_neural_network, early_stopping_scheduler.best_losses

        return early_stopping_scheduler.best_neural_network, early_stopping_scheduler.best_losses

    def multiple_restarts_training(training_data, validation_data, neural_network,
                                   target_loss=None, max_restarts=10,
                                   training_parameters={}, seed=None):
        """Algorithm that performs multiple restarts of neural network training.

        This method either performs a predefined number of restarts and returns
        the best trained network or tries to reach a given target loss and
        stops training when the target loss is reached.

        See :func:`train_neural_network` for more information on the parameters.

        Parameters
        ----------
        training_data
            Data to use during the training phase.
        validation_data
            Data to use during the validation phase.
        target_loss
            Loss to reach during training (if `None`, the network with the
            smallest loss is returned).
        max_restarts
            Maximum number of restarts to perform.
        neural_network
            The neural network to train (parameters will be reset after each
            restart).

        Returns
        -------
        best_neural_network
            The best trained neural network.
        losses
            The corresponding losses.

        Raises
        ------
        NeuralNetworkTrainingFailed
            Raised if prescribed loss can not be reached within the given number
            of restarts.
        """
        assert isinstance(training_parameters, dict)
        assert isinstance(max_restarts, int) and max_restarts > 0

        logger = getLogger('pymor.algorithms.neural_network.multiple_restarts_training')

        # if applicable, set a common seed for the PyTorch initialization
        # of weights and biases and further PyTorch methods for all training runs
        if seed:
            torch.manual_seed(seed)

        if target_loss:
            logger.info(f'Performing up to {max_restarts} restart{"s" if max_restarts > 1 else ""} '
                        f'to train a neural network with a loss below {target_loss:.3e} ...')
        else:
            logger.info(f'Performing up to {max_restarts} restart{"s" if max_restarts > 1 else ""} '
                        'to find the neural network with the lowest loss ...')

        with logger.block('Training neural network #0 ...'):
            best_neural_network, losses = train_neural_network(training_data, validation_data,
                                                               neural_network, training_parameters)

        # perform multiple restarts
        for run in range(1, max_restarts + 1):

            if target_loss and losses['full'] <= target_loss:
                logger.info(f'Finished training after {run - 1} restart{"s" if run - 1 != 1 else ""}, '
                            f'found neural network with loss of {losses["full"]:.3e} ...')
                return neural_network, losses

            with logger.block(f'Training neural network #{run} ...'):
                # reset parameters of layers to start training with a new and untrained network
                for layers in neural_network.children():
                    for layer in layers:
                        if hasattr(layer, 'reset_parameters'):
                            layer.reset_parameters()
                # perform training
                current_nn, current_losses = train_neural_network(training_data, validation_data,
                                                                  neural_network, training_parameters)

            if current_losses['full'] < losses['full']:
                logger.info(f'Found better neural network (loss of {current_losses["full"]:.3e} '
                            f'instead of {losses["full"]:.3e}) ...')
                best_neural_network = current_nn
                losses = current_losses
            else:
                logger.info(f'Rejecting neural network with loss of {current_losses["full"]:.3e} '
                            f'(instead of {losses["full"]:.3e}) ...')

        if target_loss:
            raise NeuralNetworkTrainingFailed(f'Could not find neural network with prescribed loss of '
                                              f'{target_loss:.3e} (best one found was {losses["full"]:.3e})!')
        logger.info(f'Found neural network with error of {losses["full"]:.3e} ...')
        return best_neural_network, losses



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np

from pymor.core.cache import cached
from pymor.discretizers.builtin.grids.interfaces import GridWithOrthogonalCenters
from pymor.discretizers.builtin.grids.referenceelements import triangle


class TriaGrid(GridWithOrthogonalCenters):
    r"""Basic implementation of a triangular grid on a rectangular domain.

    The global face, edge and vertex indices are given as follows ::

                 6---------10----------7---------11----------8
                 | \                 / | \                 / |
                 |    22   10    18    |    23   11    19    |
                 |       \     /       |       \     /       |
                 3    14   11     6    4    15   12     7    5
                 |       /     \       |       /     \       |
                 |    14    2    26    |    15    3    27    |
                 | /                 \ | /                 \ |
                 3----------8----------4----------9----------5
                 | \                 / | \                 / |
                 |    20    8    16    |    21    9    17    |
                 |       \     /       |       \     /       |
                 0    12    9     4    1    13   10     5    2
                 |       /     \       |       /     \       |
                 |    12    0    24    |    13    1    25    |
                 | /                 \ | /                 \ |
                 0----------6----------1----------7----------2

    Parameters
    ----------
    num_intervals
        Tuple `(n0, n1)` determining a grid with `n0` x `n1` codim-0 entities.
    domain
        Tuple `(ll, ur)` where `ll` defines the lower left and `ur` the upper right
        corner of the domain.
    identify_left_right
        If `True`, the left and right boundaries are identified, i.e. the left-most
        codim-0 entities become neighbors of the right-most codim-0 entities.
    identify_bottom_top
        If `True`, the bottom and top boundaries are identified, i.e. the bottom-most
        codim-0 entities become neighbors of the top-most codim-0 entities.
    """

    dim = 2
    reference_element = triangle

    def __init__(self, num_intervals=(2, 2), domain=([0, 0], [1, 1]),
                 identify_left_right=False, identify_bottom_top=False):
        if identify_left_right:
            assert num_intervals[0] > 1
        if identify_bottom_top:
            assert num_intervals[1] > 1
        domain = np.array(domain)
        self.__auto_init(locals())

        self.x0_num_intervals = x0_num_intervals = num_intervals[0]
        self.x1_num_intervals = x1_num_intervals = num_intervals[1]
        self.x0_range = self.domain[:, 0]
        self.x1_range = self.domain[:, 1]
        self.x0_width = self.x0_range[1] - self.x0_range[0]
        self.x1_width = self.x1_range[1] - self.x1_range[0]
        self.x0_diameter = self.x0_width / x0_num_intervals
        self.x1_diameter = self.x1_width / x1_num_intervals
        n_elements = x0_num_intervals * x1_num_intervals * 4

        # TOPOLOGY
        n_outer_vertices = (x0_num_intervals + 1 - identify_left_right) * (x1_num_intervals + 1 - identify_bottom_top)
        self.__sizes = (n_elements,
                        ((x0_num_intervals + 1 - identify_left_right) * x1_num_intervals
                         + (x1_num_intervals + 1 - identify_bottom_top) * x0_num_intervals
                         + n_elements),
                        n_outer_vertices + int(n_elements / 4))

        # calculate subentities -- codim-1
        V_EDGE_H_INDICES = np.arange(x0_num_intervals + 1, dtype=np.int32)
        if identify_left_right:
            V_EDGE_H_INDICES[-1] = 0
        V_EDGE_V_INDICES = np.arange(x1_num_intervals, dtype=np.int32) * (x0_num_intervals + 1 - identify_left_right)
        V_EDGE_INDICES = V_EDGE_V_INDICES[:, np.newaxis] + V_EDGE_H_INDICES
        num_v_edges = x1_num_intervals * (x0_num_intervals + 1 - identify_left_right)

        H_EDGE_H_INDICES = np.arange(x0_num_intervals, dtype=np.int32)
        H_EDGE_V_INDICES = np.arange(x1_num_intervals + 1, dtype=np.int32)
        if identify_bottom_top:
            H_EDGE_V_INDICES[-1] = 0
        H_EDGE_V_INDICES *= x0_num_intervals
        H_EDGE_INDICES = H_EDGE_V_INDICES[:, np.newaxis] + H_EDGE_H_INDICES + num_v_edges
        num_h_edges = x0_num_intervals * (x1_num_intervals + 1 - identify_bottom_top)

        D_EDGE_LL_INDICES = np.arange(n_elements / 4, dtype=np.int32) + (num_v_edges + num_h_edges)
        D_EDGE_UR_INDICES = D_EDGE_LL_INDICES + int(n_elements / 4)
        D_EDGE_UL_INDICES = D_EDGE_UR_INDICES + int(n_elements / 4)
        D_EDGE_LR_INDICES = D_EDGE_UL_INDICES + int(n_elements / 4)

        E0 = np.array([H_EDGE_INDICES[:-1, :].ravel(),
                       D_EDGE_LR_INDICES,
                       D_EDGE_LL_INDICES]).T
        E1 = np.array([V_EDGE_INDICES[:, 1:].ravel(),
                       D_EDGE_UR_INDICES,
                       D_EDGE_LR_INDICES]).T
        E2 = np.array([H_EDGE_INDICES[1:, :].ravel(),
                       D_EDGE_UL_INDICES,
                       D_EDGE_UR_INDICES]).T
        E3 = np.array([V_EDGE_INDICES[:, :-1].ravel(),
                       D_EDGE_LL_INDICES,
                       D_EDGE_UL_INDICES]).T

        codim1_subentities = np.vstack((E0, E1, E2, E3))

        # calculate subentities -- codim-2
        VERTEX_H_INDICES = np.arange(x0_num_intervals + 1, dtype=np.int32)
        if identify_left_right:
            VERTEX_H_INDICES[-1] = 0
        VERTEX_V_INDICES = np.arange(x1_num_intervals + 1, dtype=np.int32)
        if identify_bottom_top:
            VERTEX_V_INDICES[-1] = 0
        VERTEX_V_INDICES *= x0_num_intervals + 1 - identify_left_right
        VERTEX_NUMERS = VERTEX_V_INDICES[:, np.newaxis] + VERTEX_H_INDICES
        VERTEX_CENTER_NUMBERS = np.arange(x0_num_intervals * x1_num_intervals, dtype=np.int32) + n_outer_vertices

        V0 = np.array([VERTEX_CENTER_NUMBERS,
                       VERTEX_NUMERS[:-1, :-1].ravel(),
                       VERTEX_NUMERS[:-1, 1:].ravel()]).T
        V1 = np.array([VERTEX_CENTER_NUMBERS,
                       VERTEX_NUMERS[:-1, 1:].ravel(),
                       VERTEX_NUMERS[1:, 1:].ravel()]).T
        V2 = np.array([VERTEX_CENTER_NUMBERS,
                       VERTEX_NUMERS[1:, 1:].ravel(),
                       VERTEX_NUMERS[1:, :-1].ravel()]).T
        V3 = np.array([VERTEX_CENTER_NUMBERS,
                       VERTEX_NUMERS[1:, :-1].ravel(),
                       VERTEX_NUMERS[:-1, :-1].ravel()]).T

        codim2_subentities = np.vstack((V0, V1, V2, V3))
        self.__subentities = (codim1_subentities, codim2_subentities)

        # GEOMETRY

        # embeddings
        x0_shifts = np.arange(x0_num_intervals) * self.x0_diameter + (self.x0_range[0] + 0.5 * self.x0_diameter)
        x1_shifts = np.arange(x1_num_intervals) * self.x1_diameter + (self.x1_range[0] + 0.5 * self.x1_diameter)
        B = np.tile(np.array(np.meshgrid(x0_shifts, x1_shifts)).reshape((2, -1)).T,
                    (4, 1))

        ROT45  = np.array([[1./np.sqrt(2.),   -1./np.sqrt(2.)],
                           [1./np.sqrt(2.),    1./np.sqrt(2.)]])
        ROT135 = np.array([[-1./np.sqrt(2.),  -1./np.sqrt(2.)],
                           [1./np.sqrt(2.),   -1./np.sqrt(2.)]])
        ROT225 = np.array([[-1./np.sqrt(2.),   1./np.sqrt(2.)],
                           [-1./np.sqrt(2.),  -1./np.sqrt(2.)]])
        ROT315 = np.array([[1./np.sqrt(2.),    1./np.sqrt(2.)],
                           [-1./np.sqrt(2.),   1./np.sqrt(2.)]])
        SCAL = np.diag([self.x0_diameter / np.sqrt(2), self.x1_diameter / np.sqrt(2)])
        A0 = np.tile(SCAL.dot(ROT225), (int(n_elements / 4), 1, 1))
        A1 = np.tile(SCAL.dot(ROT315), (int(n_elements / 4), 1, 1))
        A2 = np.tile(SCAL.dot(ROT45), (int(n_elements / 4), 1, 1))
        A3 = np.tile(SCAL.dot(ROT135), (int(n_elements / 4), 1, 1))
        A = np.vstack((A0, A1, A2, A3))
        self.__embeddings = (A, B)

    def __reduce__(self):
        return (TriaGrid,
                (self.num_intervals, self.domain, self.identify_left_right, self.identify_bottom_top))

    def __str__(self):
        return (f'Tria-Grid on domain '
                f'[{self.x0_range[0]},{self.x0_range[1]}] x [{self.x1_range[0]},{self.x1_range[1]}]\n'
                f'x0-intervals: {self.x0_num_intervals}, x1-intervals: {self.x1_num_intervals}\n'
                f'elements: {self.size(0)}, edges: {self.size(1)}, vertices: {self.size(2)}')

    def size(self, codim=0):
        assert 0 <= codim <= 2, 'Invalid codimension'
        return self.__sizes[codim]

    def subentities(self, codim, subentity_codim):
        assert 0 <= codim <= 2, 'Invalid codimension'
        assert codim <= subentity_codim <= 2, 'Invalid subentity codimension'
        if codim == 0:
            if subentity_codim == 0:
                return np.arange(self.size(0), dtype='int32')[:, np.newaxis]
            else:
                return self.__subentities[subentity_codim - 1]
        else:
            return super().subentities(codim, subentity_codim)

    def embeddings(self, codim=0):
        if codim == 0:
            return self.__embeddings
        else:
            return super().embeddings(codim)

    def bounding_box(self):
        return np.array(self.domain)

    @cached
    def orthogonal_centers(self):
        embeddings = self.embeddings(0)
        ne4 = len(embeddings[0]) // 4
        if self.x0_diameter > self.x1_diameter:
            x0_fac = (self.x1_diameter / 2) ** 2 / (3 * (self.x0_diameter / 2) ** 2)
            x1_fac = 1./3.
        else:
            x1_fac = (self.x0_diameter / 2) ** 2 / (3 * (self.x1_diameter / 2) ** 2)
            x0_fac = 1./3.
        C0 = embeddings[0][:ne4].dot(np.array([x1_fac, x1_fac])) + embeddings[1][:ne4]
        C1 = embeddings[0][ne4:2*ne4].dot(np.array([x0_fac, x0_fac])) + embeddings[1][ne4:2*ne4]
        C2 = embeddings[0][2*ne4:3*ne4].dot(np.array([x1_fac, x1_fac])) + embeddings[1][2*ne4:3*ne4]
        C3 = embeddings[0][3*ne4:4*ne4].dot(np.array([x0_fac, x0_fac])) + embeddings[1][3*ne4:4*ne4]
        return np.concatenate((C0, C1, C2, C3), axis=0)

    def visualize(self, U, codim=2, **kwargs):
        """Visualize scalar data associated to the grid as a patch plot.

        Parameters
        ----------
        U
            |NumPy array| of the data to visualize. If `U.dim == 2 and len(U) > 1`, the
            data is visualized as a time series of plots. Alternatively, a tuple of
            |Numpy arrays| can be provided, in which case a subplot is created for
            each entry of the tuple. The lengths of all arrays have to agree.
        codim
            The codimension of the entities the data in `U` is attached to (either 0 or 2).
        kwargs
            See :func:`~pymor.discretizers.builtin.gui.visualizers.PatchVisualizer.visualize`
        """
        from pymor.discretizers.builtin.gui.visualizers import PatchVisualizer
        from pymor.vectorarrays.interface import VectorArray
        from pymor.vectorarrays.numpy import NumpyVectorSpace, NumpyVectorArray
        if isinstance(U, (np.ndarray, VectorArray)):
            U = (U,)
        assert all(isinstance(u, (np.ndarray, VectorArray)) for u in U)
        U = tuple(NumpyVectorSpace.make_array(u) if isinstance(u, np.ndarray) else
                  u if isinstance(u, NumpyVectorArray) else
                  NumpyVectorSpace.make_array(u.to_numpy())
                  for u in U)
        PatchVisualizer(self, codim=codim).visualize(U, **kwargs)
